/*****************************************************************************
    getarg.c

    Diese Datei ist Teil des Programmpakets 'ACL-tftp'.
    Alle Rechte bei Andreas G. Schindler
                    Klappacher Str. 2i
                    64285 Darmstadt, Deutschland.

    Dieses Programm ist 'Freie Software'; Sie knnen es entsprechend den 
    Bedingungen der 'Allgemeinen ffentlichen GNU-Lizenz' (verffentlicht 
    durch die 'Free Software Foundation') weiter verbreiten und/oder 
    modifizieren. Massgeblich ist die Version 2 oder eine neuere Version
    dieser Lizenz. 

    Dieses Programm wird in der Hoffnung dass es ntzlich is weitergegeben,
    jedoch OHNE JEDE GARANTIE. Dieser Garantieauschluss bezieht sich auch
    auf Eigenschaften wie Marktfhigkeit oder Eignung zu einem bestimmten
    Zweck. Details finden sich in der 'Allgemeinen ffentlichen GNU-Lizenz'.

    Eine Kopie dieser Lizenz sollte in  Programmpaket, dem diese Datei
    entstammt, enthalten sein. Falls nicht, schreiben Sie an die 
    'Free Software Foundation, Inc.', 59 Temple Place, Suite 330, Boston, 
    MA  02111-1307  USA.


    This file is part of the 'ACL-tftp' Package
    Copyright (C) 1999-2005 by Andreas G. Schindler
                               Klappacher Str. 2i
                               Darmstadt, Germany.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************************************************/
#include "auto.h"
#include <stdio.h>
#include <string.h>

#include "params.h"
#include "getarg.h"
/*
    getarg --- Parse Argument line

    Parameters: argc	int	Argument count, elements of argv[], 
				as given by main()

		argv[]	char**	Array of strings containing the broken dowwm
				command line argv[0] contains the program name.

		allopts	char*	String containing all option characters
				without leading '-'. Options that may comprise
				a value must be followed by a colon.
				Example: "abc:de:f"
                                Options not listed in allopts are treated as
				value-less options.

		opt		Option to scan for (character) or a negative
				integer value starting with -1 to scan for 
				an argument.

		defval	char*	value to be returned if an option is specified
				without a value. Meaningless with parameters

		notfound char*	value to be returned if the option/parameter
				is not found during the scan. Returned also
				if a requested parameter is not present.

		flag int*	address of a flag to examine what to do.
				this flag will be modified to reflect the
				action performed.
*/

char *getArg (int argc, char *argv[], char *allopts, int opt, 
              char *defval, char *notfound, int *set)
{
    int i;
    char *ptr, *otv;

    if (opt >= 0) {
        if ((otv=strchr(allopts,opt)) == NULL) 
            otv = "";
        else
            otv++;

        for (i = 1; i < argc; i++) {
            ptr = argv[i];
            if (*ptr++ == '-') {
                if (*ptr == '\0') continue;	/* dash-only ignored */
                if (*ptr++ == (char)opt) {
                    if (*otv != ':') {
                        if (set != NULL) *set = GA_DEFAULT;
                        return (defval);
                    }
                    if (*ptr == '\0') {
                        if (i == (argc-1)) {
                            if (set != NULL) *set = GA_DEFAULT;
                            return (defval);
                        }
                        else {
                            ptr = argv[i+1];
                            if (*ptr == '-') {
                                if (set != NULL) *set = GA_DEFAULT;
                                return (defval);
                            }
                            else {
                                if (set != NULL) *set = GA_SETVAL;
                                return (ptr);
                            }
                        }
                    }
                    else {
                        if (set != NULL) *set = GA_SETVAL;
                        return (ptr);
                    }
                }
            }
        }
        if (set != NULL) *set = GA_NOTFOUND;
        return (notfound);
    }
    else {
        int last;

        for (i = 1, last = 0; i < argc; i++) {
            ptr = argv[i];
            if (*ptr++ != '-') break;	/* option list stops with first argument */
            if (*ptr == '\0') break;	/* dash-only treated as an argument */
            last = i;
            if ((otv=strchr(allopts,*ptr++)) != NULL)
                if (*(++otv) == ':')
                    if (*ptr == '\0')
                        if (i < (argc-1))
                            if (*argv[i+1] != '-') last++;
        }
        if (last-opt >= argc) {
            if (set != NULL) *set = GA_NOTFOUND;
            return (notfound);
        }
        if (set != NULL) *set = GA_SETVAL;
        return (argv[last-opt]);
    }
}

char *getarg (int argc, char *argv[], char *allopts, int opt, 
              char *defval, char *notfound)
{
    return getArg (argc, argv, allopts, opt, defval, notfound, NULL);
}

#ifdef TEST
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "util.h"

#ifdef DEBUG

int main (int argc, char *argv[])
{
    char *p1, *p2, *p3, *ctx2, *ctx3, *a1, *a2, *a3, *b1, *b2, *b3;

    if (argc < 3) exit (0);

    a1 = b1 = (char *)Malloc(strlen(argv[1])+1);
    a2 = b2 = (char *)Malloc(strlen(argv[1])+1);
    a3 = b3 = (char *)Malloc(strlen(argv[1])+1);
    strcpy(a1, argv[1]);
    strcpy(a2, argv[1]);
    strcpy(a3, argv[1]);

    do {
        p1 =  strtok (a1,argv[2]);
        p2 = xstrtok (a2,argv[2],&ctx2);
        p3 = zstrtok (a3,argv[2],&ctx3);
        printf(" strtok: [%s] [%s]\n"
               "xstrtok: [%s] [%s]\n"
               "zstrtok: [%s] [%s]\n",p1,b1,p2,b2,p3,b3);
        a1 = a2 = a3 = NULL;
    } while (p1 != NULL || p2 != NULL || p3 != NULL);

    return (0);
}
#endif
static char **slice (char *alist)
{
    char *ptr, *ctx;
    char **res;
    int  numargs = 0;
#define res0size 32

    for (ptr  = zstrtok(alist,"\t ",&ctx); 
         ptr != NULL;
         ptr  = zstrtok(NULL,"\t ",&ctx)) {
        Free(ptr);
        numargs++;
    }

    if ((res = (char **)Malloc(sizeof(char **)*(numargs+1))) == NULL) 
        return (NULL);

    if ((res[0] = (char *)Malloc(res0size)) == NULL) 
        return (NULL);

    snprintf(res[0], res0size, "%d",numargs+1);

    numargs = 1;

    for (ptr  = xstrtok(alist,"\t ",&ctx);
         ptr != NULL;
         ptr = xstrtok(NULL,"\t ",&ctx)) {
        res[numargs++] = ptr;
    }
    return(res);
#undef res0size
}


int main (int argc, char *argv[])
{
    char **av, *res;
    int na, i;

    if (argc < 6) {
        puts("Insufficient number of arguments (< 5)");
        exit (1);
    }

    av = slice(argv[1]);

    printf("%d arguments in test string ",atoi(av[0])-1);
    for (i=1; i<atoi(av[0]); i++) printf("[%s] ",av[i]);
    printf("\nOption definition string is [%s]\n",argv[2]);
    if (sscanf(argv[3],"%d",&na) == 1 && na < 0) {
        printf("Looking for argument [%d], default: [%s]\n",-na,argv[5]);
        res = getarg(atoi(av[0]), av, argv[2], na      , argv[4], argv[5]);
    }
    else {
        printf("Looking for option [%c], defaults: [%s] [%s]\n",
               *argv[3],argv[4],argv[5]);
        res = getarg(atoi(av[0]), av, argv[2], *argv[3], argv[4], argv[5]);
    }
    if (strlen(res)) 
        puts(res);

    return(0);
}
#endif
