/*****************************************************************************
    main.c

    Diese Datei ist Teil des Programmpakets 'ACL-tftp'.
    Alle Rechte bei Andreas G. Schindler
                    Klappacher Str. 2i
                    64285 Darmstadt, Deutschland.

    Dieses Programm ist 'Freie Software'; Sie knnen es entsprechend den 
    Bedingungen der 'Allgemeinen ffentlichen GNU-Lizenz' (verffentlicht 
    durch die 'Free Software Foundation') weiter verbreiten und/oder 
    modifizieren. Massgeblich ist die Version 2 oder eine neuere Version
    dieser Lizenz. 

    Dieses Programm wird in der Hoffnung dass es ntzlich is weitergegeben,
    jedoch OHNE JEDE GARANTIE. Dieser Garantieauschluss bezieht sich auch
    auf Eigenschaften wie Marktfhigkeit oder Eignung zu einem bestimmten
    Zweck. Details finden sich in der 'Allgemeinen ffentlichen GNU-Lizenz'.

    Eine Kopie dieser Lizenz sollte in  Programmpaket, dem diese Datei
    entstammt, enthalten sein. Falls nicht, schreiben Sie an die 
    'Free Software Foundation, Inc.', 59 Temple Place, Suite 330, Boston, 
    MA  02111-1307  USA.


    This file is part of the 'ACL-tftp' Package
    Copyright (C) 1999-2005 by Andreas G. Schindler
                               Klappacher Str. 2i
                               Darmstadt, Germany.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************************************************/
#include "auto.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <signal.h>
#include <sys/ipc.h>
#include <sys/sem.h>

#include "globals.h"
#include "access.h"
#include "readconfig.h"
#include "fileio.h"
#include "netio.h"
#include "logfun.h"
#include "process.h"
#include "util.h"
#include "dump.h"

/*
	select the appropriate transfer type and process it
*/

static int processRequest2 (sopair_t *sockets, tftpR_t *req)
{
    int rtc;

    if ((rtc = validateAccess(req))) {
        if (!silentDeny && !(rtc & PX_DISCARD)) {
            req->reqSock = *sockets;
            sendError(req, (rtc & PE_ERRMASK), NULL);
        }
        return 1;
    }

    if (connectPeer(req))
        return 1;

    if(req->times.eflag != SD_UNSET)	/* disable abort tftp message */
        silentDeny = req->times.eflag;

    switch (req->reqType) {
        case RT_RRQ:
            rtc = sendFile(req);
            break;
        case RT_WRQ:
            rtc = recvFile(req);
            break;
        default:
            rtc = 1;
            break;
    }

    if (!rtc)
        DO_MSG3 statusTrace (req);

    return rtc;
}

/*
	jacket routine for processRequest2()
*/

int processRequest (sopair_t *sockets, tftpR_t *req)
{
    int rtc;

    regCon.req    = (struct tftpR_s *)req;
    regCon.pid    = (int)getpid();

    rtc = processRequest2 (sockets, req);

    regCon.pid    = 0;
    regCon.req    = NULL;

    close (req->reqFd);    
    close (req->Rsock);    
    close (req->Ssock);    
    return rtc;
}


/*
	main TFTPD server program

*/

/*	Note:	lines marked 					*/
				/**/ 
/*					run in a child process 	*/	 

int main (int argc, char **argv)
{
    tftpR_t    *req;			/* request data			*/
    sopair_t   *sockets;		/* primary sockets		*/
    sopair_t   ftsocks;			/* pr/socks if started by inetd */
    int        children;		/* currently active children	*/
    int        rtc;			/* getRequest return code	*/

/*
	Startup-Phase
	parse the command line input and set the globals consitently
	the configuration file is read & interpreted for the 1st time
	later, at the begin of each transfer, the configuration file
	is checked for any modigfications and re-parsed if necessary
	if only a configuration check is launched, we're finished here
*/
    getcwd(defaultDir, sizeof(defaultDir));
    initLocale("");

#ifdef MALLOC_TRACE
    mtrace();
#endif

    parentPid = getpid();		/* reflect our new status 	*/

    atexit (processRundown);
    if (setjmp(abortBuf))
        exit (1);

    evalCommandLine (argc, argv, 1);	/* get command line params	*/

    startLog(INIT_LOGGER);		/* primary logging setup	*/

    rtc = readconfigFile(1);		/* parse tconfig file		*/

    startLog(UPDATE_LOGGER);		/* primary logging setup	*/

#ifdef FILE_CACHE

    preloadCache();
    if (checkOnly)
        cacheDump(stderr, 3);

#endif /* FILE_CACHE */


    if (checkOnly) {
        if (rtc) {
            LOG_MSG1 (LOG_ERR, _("Configuration check returns %d errors"), rtc);
        }
        else {
            LOG_MSG4 (LOG_INFO, _("Configuration checks o.k."));
        }
        exit (rtc);
    }
    else {
        if (rtc) {
            LOG_MSG1 (LOG_ERR, _("%d configuration errors detected"), rtc);
            exit (1);
        }
        else {
            DO_MSG4 {
                logBuild();
                doLog (LOG_INFO, _("working directory '%s'"), defaultDir);
                doLog (LOG_INFO, _("local bindings: %s:%d"),
                       xinet_ntoa(serverAddr,1), serverPort);
                if (allowRecover)
                    doLog (LOG_INFO, _("RRQ recovery enabled"));

            }
        }
    }

    if (accTest != NULL)		/* access test only	*/
        return (validateAccess(req));

    if (checkPidFile()) {
        LOG_MSG1 (LOG_ERR, _("PID file exists - quitting"));
        exit (1);
    }

/*
	now we're sure there's some real work ahead
	we must decide, wich mode of operation should be entered
		- started by inetd,
		- standalone daemon or 
		- standalone / single-process (debug mode)
*/

/*
	here on single process mode for debugging
	we run in an endless loop, processing all request serialized
*/
    if (spMode) {
						/* initialize network	*/
        if((sockets=netStart(serverAddr, serverPort))==NULL)
            exit (1);

        if (signalSetup(PAR_SET,0))
            exit (1);

        for (;;) {

            startTimer(&idleTimer);

            while (!checkTimer(idleTimer, defTimer.idle)) {

                if (readconfigFile(0)) 	/* check for modifications	*/
                    exit (1);

                if (doDump) {
                  doDump = 0;
                  genDump(3);
                }

                rtc = getRequest(sockets, &req);

                if (rtc < 0)
                    exit (1);
                else if (rtc > 0) 
                    break;
            }

            if (!rtc) {
                LOG_MSG3 (LOG_NOTICE, _("Exiting after %ld seconds idle"),
                          defTimer.idle.tv_sec);
                break;
            }

            changeIdent();		/* change uid, gid & root	*/
            processRequest(sockets, req);	/* do the final work	*/
            flushRequest(&req);		/* dismiss request storage	*/
            resetIdent();		/* switch back to parent u/gid	*/
        }
    }
/*
	this is the 'daemon' section
	each new transfer request causes the spawn of a new child process
	the child then changes UID/GID and the root if desired
	from this new (low privileged) incarnation of tftpd the access 
	validation takes place and, if successful, the transfer is done
*/
 
    else {

        if (isaSocket(stdin)) {		/* started by inetd ?		*/
            sockets = &ftsocks;
            sockets->recv = fileno(stdin);
            sockets->send = fileno(stdout);
            if (signalSetup(PAR_SET,0))
                exit (1);
        }
        else {				/* no, we're standalone		*/

            if (noDetach) {
                if (procEnvSetup())
                    exit (1);
            }
            else {
                if (becomeDaemon()) 
                    exit (1);
            }

            if (signalSetup(PAR_SET,0))
                exit (1);

            if ((sockets = netStart(serverAddr, 
                                    serverPort)) == NULL)  /* init network */
                exit (1);
        }

        SIGBLOCK(SIGCHLD, (startTimer(&idleTimer)));

        for(;;) {

            for (;;) {
                SIGBLOCK(SIGCHLD, (rtc=checkTimer(idleTimer, defTimer.idle)));
/* 
	the total idle time exceeds the maximum idle time, so we're to stop. 
	Additionally, all children so far must have finished execution.
*/
                if (rtc) {
                    SIGBLOCK(SIGCHLD, (children = getPidCount(procListH)));
                    if (!children) {
                        LOG_MSG3 (LOG_NOTICE, 
                                  _("Exiting after %ld seconds idle"),
                                  defTimer.idle.tv_sec);
                        exit(0);
                    }
                    LOG_MSG4 (LOG_INFO,
                              _p("%d child busy", "%d children busy", children),
                              children);
                    DO_MSG3 statsRequest();
                    SIGBLOCK(SIGCHLD, (startTimer(&idleTimer)));
                }

                if (readconfigFile(0))	/* check for modifications	*/
                    exit (1);

                if (doDump) {
                  doDump = 0;
                  genDump(3);
                }

                rtc = getRequest(sockets, &req);

		LOG_MSG5(LOG_DEBUG,"getRequest() returns %d", rtc);

                if (rtc < 0)		/* severe error condition	*/
                    exit (1);

                if (rtc > 0) {		/* xfer request pending		*/
                    SIGBLOCK(SIGCHLD, (startTimer(&idleTimer)));
                    break;
                }
            }
/*
	getRequest() shoots up a legal xfer-request
*/
            if (sSpawn(req->limGroup, &req->reqPeer)) {	/* parent	*/
                flushRequest(&req);	/* dismiss request storage	*/
                continue;		/* and continue 		*/
            }
      /**/  deletePidChain(procListH);	/* proc list starts allover	*/
      /**/  startLog(FULL_LOGGER);	/* assure our own log channel	*/
      /**/  changeIdent();		/* change uid, gid & root	*/
      /**/
      /**/  rtc = processRequest(sockets, req); /* do the real work	*/
      /**/  
      /**/  exit (rtc);
        }
    } 
    return rtc;
}
